(******************************************************************************)
(* computer (stack machine)                                                   *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Comp;

FROM SYSTEM IMPORT ADR, BYTE;
IMPORT Lib, FIO;
FROM Comp0 IMPORT TValue, TIndex, TOperation, TFunction;
FROM Comp0 IMPORT TCmdDesc, TSet, TCmdStr;
FROM Comp0 IMPORT acMax, vcMax, pcMax;
FROM Comp0 IMPORT Convert;

CONST
  debug = TRUE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

VAR
  (* program *)
  prg: ARRAY [ 0 .. pcMax - 1 ] OF BYTE;
  pos, len: TAddress;
  (* machine state (execution) *)
  cmd: TCmdDesc;
  pc, pc0: TAddress;
  args: ARRAY [ 1 .. acMax ] OF TValue;  (* stack *)
  ac: CARDINAL;
  vars: ARRAY [ 0 .. vcMax - 1 ] OF TValue;  (* data *)
  cond: BOOLEAN;
  (* module state *)
  strs: ARRAY BOOLEAN OF ARRAY [ 0 .. 4 ] OF CHAR;
  first: BOOLEAN;

(***** debugging *****)

PROCEDURE Dbg();
VAR s: TCmdStr;
BEGIN
  Convert( cmd, s );
  IF first THEN
    FIO.WrStr( dbg, 'Comp.Dbg:   pc  cond  ac       top-1         top' );
    FIO.WrLn( dbg );
    first := FALSE;
  END;
  FIO.WrStr( dbg, 'Comp.Dbg: ' );
  FIO.WrLngCard( dbg, pc0, 4 );
  FIO.WrStr( dbg, ': ' );
  FIO.WrStr( dbg, strs[cond] );
  FIO.WrChar( dbg, ' ' );
  FIO.WrLngCard( dbg, ac, 2 );
  FIO.WrChar( dbg, ' ' );
  IF ac = 0 THEN
    FIO.WrStr( dbg, '                       ' );
  ELSIF ac = 1 THEN
    FIO.WrStr( dbg, '            ' );
    FIO.WrLngInt( dbg, args[ac], 11 );
  ELSE
    FIO.WrLngInt( dbg, args[ac-1], 11 );
    FIO.WrChar( dbg, ' ' );
    FIO.WrLngInt( dbg, args[ac], 11 );
  END;
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
  FIO.WrLn( dbg );
END Dbg;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE ErrCard( s: ARRAY OF CHAR; x: CARDINAL );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrChar( err, ' ' );
  FIO.WrLngCard( err, x, 0 );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END ErrCard;

(***** programming *****)

PROCEDURE Write( x-: ARRAY OF BYTE );
VAR i: CARDINAL;
BEGIN
  IF pos + SIZE( x ) > pcMax THEN Err( 'Program too large' ) END;
  FOR i := 0 TO HIGH( x ) DO prg[ pos ] := x[ i ]; INC( pos ) END;
  IF pos > len THEN len := pos END;
END Write;

PROCEDURE GetPos( VAR x: TAddress );
BEGIN
  x := pos;
END GetPos;

PROCEDURE SetPos( x: TAddress );
BEGIN
  IF x > len THEN Err( 'Invalid position' ) END;
  pos := x;
END SetPos;

PROCEDURE Read( VAR x: ARRAY OF BYTE );
VAR i: CARDINAL;
BEGIN
  IF pos + SIZE( x ) > len THEN Err( 'Nothing to read' ) END;
  FOR i := 0 TO HIGH( x ) DO x[ i ] := prg[ pos ]; INC( pos ) END;
END Read;

(***** execution / listing *****)

PROCEDURE FetchCmd();

  PROCEDURE Read( VAR x: ARRAY OF BYTE );
  VAR i: CARDINAL;
  BEGIN
    IF pc + SIZE( x ) > len THEN Err( 'Bug: End of program' ) END;
    FOR i := 0 TO HIGH( x ) DO x[ i ] := prg[ pc ]; INC( pc ) END;
  END Read;

BEGIN
  pc0 := pc;
  Read( cmd.fct );
  CASE cmd.fct OF
  | fHlt      :
  | fVal      : Read( cmd.val );
  | fLod, fSto: Read( cmd.idx );
  | fOpr      : Read( cmd.opr );
  | fJmp..fJpt: Read( cmd.adr );
  ELSE
    ErrCard( 'Invalid function', ORD(cmd.fct) );
  END;
END FetchCmd;

PROCEDURE ChkIdx();
BEGIN
  IF cmd.idx >= vcMax THEN ErrCard( 'Invalid index', cmd.idx ) END;
END ChkIdx;

PROCEDURE ChkAdr();
BEGIN
  IF cmd.adr > len THEN ErrCard( 'Invalid address', cmd.adr ) END;
END ChkAdr;

PROCEDURE Push( x: TValue );
BEGIN
  IF ac >= acMax THEN Err( 'Stack overflow' ) END;
  INC( ac ); args[ ac ] := x;
END Push;

PROCEDURE Pop( VAR x: TValue );
BEGIN
  IF ac <= 0 THEN Err( 'Stack underflow' ) END;
  x := args[ ac ]; DEC( ac );
END Pop;

PROCEDURE ExecCmd();
VAR x,y: TValue;
BEGIN
  CASE cmd.fct OF
  | fHlt: pc := len;
  | fVal: Push( cmd.val );
  | fLod: ChkIdx(); Push( vars[cmd.idx] );
  | fSto: ChkIdx(); Pop( vars[cmd.idx] );
  | fOpr: CASE cmd.opr OF
          | oWrt: Pop( x ); FIO.WrLngInt( out, x, 0 ); FIO.WrLn( out );
          | oWrC: FIO.WrStr( out, strs[cond] ); FIO.WrLn( out );
          | oNeg: Pop( x ); Push( -x );
          | oAdd: Pop( y ); Pop( x ); Push( x + y );
          | oSub: Pop( y ); Pop( x ); Push( x - y );
          | oMul: Pop( y ); Pop( x ); Push( x * y );
          | oDiv: Pop( y ); Pop( x );
                  IF y = 0 THEN Err( 'Division by zero' ) END;
                  Push( x DIV y );
          | oBus: Pop( y ); Pop( x ); Push( y - x );
          | oVid: Pop( y ); Pop( x );
                  IF x = 0 THEN Err( 'Division by zero' ) END;
                  Push( y DIV x );
          | oEql: Pop( y ); Pop( x ); cond := x = y;
          | oNeq: Pop( y ); Pop( x ); cond := x # y;
          | oLss: Pop( y ); Pop( x ); cond := x < y;
          | oGtr: Pop( y ); Pop( x ); cond := x > y;
          | oLeq: Pop( y ); Pop( x ); cond := x <= y;
          | oGeq: Pop( y ); Pop( x ); cond := x >= y;
          | oNot: Pop( x ); Push( TValue(TSet(x)/TSet(TValue(-1))) );
          | oOr : Pop( y ); Pop( x ); Push( TValue(TSet(x)+TSet(y)) );
          | oAnd: Pop( y ); Pop( x ); Push( TValue(TSet(x)*TSet(y)) );
          ELSE
            ErrCard( 'Invalid operation', ORD(cmd.opr) );
          END;
  | fJmp: ChkAdr(); pc := cmd.adr;
  | fJpf: ChkAdr(); IF ~ cond THEN pc := cmd.adr END;
  | fJpt: ChkAdr(); IF  cond  THEN pc := cmd.adr END;
  ELSE
    ErrCard( 'Invalid function', ORD(cmd.fct) );
  END;
END ExecCmd;

PROCEDURE Exec();
BEGIN
  pc := 0;
  WHILE pc < len DO
    FetchCmd();
    IF debug THEN Dbg() END;
    ExecCmd();
  END;
END Exec;

PROCEDURE List();
VAR s: TCmdStr;
BEGIN
  pc := 0;
  WHILE pc < len DO
    FetchCmd();
    Convert( cmd, s );
    FIO.WrLngCard( out, pc0, 4 );
    FIO.WrStr( out, ': ' );
    FIO.WrStr( out, s );
    FIO.WrLn( out );
  END;
END List;

BEGIN
  (* program *)
  Lib.Fill( ADR(prg), SIZE(prg), fHlt );
  pos := 0; len := 0;
  (* machine state *)
  pc := 0; pc0 := 0;
  Lib.Fill( ADR(args), SIZE(args), 0C );
  ac := 0;
  Lib.Fill( ADR(vars), SIZE(vars), 0C );
  cond := FALSE;
  (* module state *)
  strs[ FALSE ] := 'FALSE';
  strs[ TRUE  ] := 'TRUE ';
  first := TRUE;
END Comp.
